
__all__ = ["AutoEncoder",
           "VariationalAutoEncoder",
           "GaussianMixtureVariationalAutoEncoder",
           #    "GNNAutoEncoder",
           "CNNAutoEncoder",
           "SPIB"]

from .ae import AutoEncoder
from .vae import VariationalAutoEncoder
from .gmvae import GaussianMixtureVariationalAutoEncoder
from .molearn import CNNAutoEncoder
from .spib import SPIB
# from .gnnae import GNNAutoEncoder
